﻿// Climate Framework for Uncertainty, Negotiation and Distribution
// Version 3.5
// (c) David Anthoff and Richard S.J. Tol
// http://www.fund-model.org
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;
using System.IO;
using System.Globalization;
using System.Data.OleDb;
using System.Data;

namespace Fund
{
    public class Parameters
    {
        private Dictionary<ParameterKey, Parameter> _parameters = new Dictionary<ParameterKey, Parameter>();
        private Dictionary<string, int> m_allParameterIdsByName = new Dictionary<string, int>();

        protected void Add(Parameter Parameter)
        {
            if (_parameters.ContainsKey(Parameter.Key))
                _parameters[Parameter.Key] = Parameter;
            else
                _parameters.Add(Parameter.Key, Parameter);
        }

        protected void Add(ParameterKey ParameterKey, string ValueString)
        {
            string i_value = String.Empty;
            double i_fake = 0.0;

            i_value = ValueString.Trim();

            // TODO Control, whether NumberStyles.Float is appropriate
            if (!(Double.TryParse(i_value, NumberStyles.Float, CultureInfo.InstalledUICulture, out i_fake) || (Double.TryParse(i_value, NumberStyles.Float, CultureInfo.InvariantCulture, out i_fake))))
            {
                if (i_value.StartsWith(c_normalDistributionName + "(") && i_value.EndsWith(")"))
                {
                    i_value = i_value.Substring(c_normalDistributionName.Length + 1, i_value.Length - (c_normalDistributionName.Length + 2));
                    Add(new NormalDistributionParameter(ParameterKey, i_value.Split(';'), _rand));
                }
                else if (i_value.StartsWith(c_triangularDistributionName + "(") && i_value.EndsWith(")"))
                {
                    i_value = i_value.Substring(c_triangularDistributionName.Length + 1, i_value.Length - (c_triangularDistributionName.Length + 2));
                    Add(new TriangularDistributionParameter(ParameterKey, i_value.Split(';'), _rand));
                }
                else if (i_value.StartsWith(c_exponentialDistributionName + "(") && i_value.EndsWith(")"))
                {
                    i_value = i_value.Substring(c_exponentialDistributionName.Length + 1, i_value.Length - (c_exponentialDistributionName.Length + 2));
                    Add(new ExponentialDistributionParameter(ParameterKey, i_value.Split(';'), _rand));
                }
                else if (i_value.StartsWith(c_gammaDistributionName + "(") && i_value.EndsWith(")"))
                {
                    i_value = i_value.Substring(c_gammaDistributionName.Length + 1, i_value.Length - (c_gammaDistributionName.Length + 2));
                    Add(new GammaDistributionParameter(ParameterKey, i_value.Split(';'), _rand));
                }
                else
                    throw new ApplicationException(string.Format("Syntax error: '{0}' is not a valid value for {1}.", i_value, ParameterKey.Name));
            }
            else
            {
                // TODO Control, whether NumberStyles.Float is appropriate
                if (Double.TryParse(i_value, NumberStyles.Float, CultureInfo.InstalledUICulture, out i_fake))
                    Add(new ConstantParameter(ParameterKey, Convert.ToDouble(i_value, CultureInfo.InstalledUICulture)));
                else
                    Add(new ConstantParameter(ParameterKey, Convert.ToDouble(i_value, CultureInfo.InvariantCulture)));
            }
        }

        protected void UseReflectionToSetFieldsCurrentValues()
        {
            System.Type i_Type;
            System.Reflection.FieldInfo[] i_fields;
            int i, l, k;
            object[] i_attributes;
            RegionParameter i_RegionParameter;
            YearParameter i_YearParameter;
            YearRegionParameter i_YearRegionParameter;

            i_Type = this.GetType();
            i_fields = i_Type.GetFields();

            for (i = 0; i < i_fields.Length; i++)
            {
                i_attributes = i_fields[i].GetCustomAttributes(typeof(ParameterAttribute), false);

                if (i_attributes.Length == 1)
                {

                    string parameterName = ((ParameterAttribute)i_attributes[0]).Name == null ? i_fields[i].Name : ((ParameterAttribute)i_attributes[0]).Name;

                    if (i_fields[i].FieldType.Name == "Double")
                    {
                        i_fields[i].SetValue(this, _parameters[new GlobalParameterKey(parameterName)].Value);
                    }
                    else if (i_fields[i].FieldType.Name == "RegionParameter")
                    {
                        i_RegionParameter = ((RegionParameter)i_fields[i].GetValue(this));
                        for (l = 1; l <= LegacyParameterConstants.NoReg; l++)
                            i_RegionParameter[l] = _parameters[new RegionParameterKey(parameterName, l)].Value;
                    }
                    else if (i_fields[i].FieldType.Name == "RegionRegionParameter")
                    {
                        var i_RegionRegionParameter = ((RegionRegionParameter)i_fields[i].GetValue(this));
                        for (l = 1; l <= LegacyParameterConstants.NoReg; l++)
                            for (k = 1; k <= LegacyParameterConstants.NoReg; k++)
                                i_RegionRegionParameter[l, k] = _parameters[new RegionRegionParameterKey(parameterName, l, k)].Value;
                    }
                    else if (i_fields[i].FieldType.Name == "YearParameter")
                    {
                        i_YearParameter = ((YearParameter)i_fields[i].GetValue(this));
                        for (l = 0; l <= LegacyParameterConstants.NYear; l++)
                            i_YearParameter[l] = _parameters[new YearParameterKey(parameterName, l)].Value;
                    }
                    else if (i_fields[i].FieldType.Name == "YearRegionParameter")
                    {
                        i_YearRegionParameter = ((YearRegionParameter)i_fields[i].GetValue(this));
                            for (l = 0; l <= LegacyParameterConstants.NYear; l++)
                                for (k = 1; k <= LegacyParameterConstants.NoReg; k++)
                                    i_YearRegionParameter[l, k] = _parameters[new YearRegionParameterKey(parameterName, l, k)].Value;
                    };
                }
                else if (i_attributes.Length > 1)
                    throw new Exception(String.Format("More than one ParameterAttribute specified for '{0}'.", i_fields[i].Name));
            }
        }

        protected void UseReflectionToCreateObjects()
        {
            System.Type i_Type;
            System.Reflection.FieldInfo[] i_fields;
            int i;
            object[] i_attributes;

            i_Type = this.GetType();
            i_fields = i_Type.GetFields();

            for (i = 0; i < i_fields.Length; i++)
            {
                i_attributes = i_fields[i].GetCustomAttributes(typeof(ParameterAttribute), false);
                if (i_attributes.Length == 1)
                {
                    if (i_fields[i].FieldType.Name == "RegionParameter")
                    {
                        i_fields[i].SetValue(this, new RegionParameter());
                    }
                    else if (i_fields[i].FieldType.Name == "RegionRegionParameter")
                    {
                        i_fields[i].SetValue(this, new RegionRegionParameter());
                    }
                    else if (i_fields[i].FieldType.Name == "YearParameter")
                    {
                        i_fields[i].SetValue(this, new YearParameter());
                    }
                    else if (i_fields[i].FieldType.Name == "YearRegionParameter")
                    {
                        i_fields[i].SetValue(this, new YearRegionParameter());
                    };
                }
                else if (i_attributes.Length > 1)
                    throw new Exception(String.Format("More than one ParameterAttribute specified for '{0}'.", i_fields[i].Name));
            }
        }

        protected void ReadExcel2003File(string Filename)
        {
            throw new NotImplementedException();
        }

        protected void ReadExcel2007File(string Filename)
        {
            int i;
            int l;
            int lParameterId = 0;

            try
            {
                Console.Write("Loading parameters from Excel file");

                var excelParameters = new Esmf.FundExcelParameterFile(Filename);
                excelParameters.Load();

                foreach (Esmf.FundExcelParameterFile.Parameter p in excelParameters.Parameters)
                {
                    int parameterId = -1;

                    if (!m_allParameterIdsByName.TryGetValue(p.Name, out parameterId))
                    {
                        parameterId = lParameterId;
                        m_allParameterIdsByName.Add(p.Name, parameterId);
                        lParameterId++;
                    }

                    if (p is Esmf.FundExcelParameterFile.NonDimensionalParameter)
                    {
                        var pp = (Esmf.FundExcelParameterFile.NonDimensionalParameter)p;
                        Add(new GlobalParameterKey(p.Name), pp.Value);
                    }
                    else if (p is Esmf.FundExcelParameterFile.OneDimensionalParameter)
                    {
                        var pp = (Esmf.FundExcelParameterFile.OneDimensionalParameter)p;

                        if (pp.Length == 16)
                        {
                            for (i = 1; i <= LegacyParameterConstants.NoReg; i++)
                            {
                                Add(new RegionParameterKey(p.Name, i), pp[i - 1]);
                            };
                        }
                        else
                        {
                            for (i = 0; i <= LegacyParameterConstants.NYear; i++)
                            {
                                Add(new YearParameterKey(p.Name, i), pp[i]);
                            };
                        };
                    }
                    else if (p is Esmf.FundExcelParameterFile.TwoDimensionalParameter)
                    {
                        var pp = (Esmf.FundExcelParameterFile.TwoDimensionalParameter)p;

                        if (pp.RowCount > 16)
                        {
                            for (i = 1; i <= LegacyParameterConstants.NoReg; i++)
                            {
                                for (l = 0; l <= LegacyParameterConstants.NYear; l++)
                                {
                                    Add(new YearRegionParameterKey(p.Name, l, i), pp[i - 1, l]);
                                };
                            };

                        }
                        else
                        {
                            for (i = 1; i <= LegacyParameterConstants.NoReg; i++)
                            {
                                for (l = 1; l <= LegacyParameterConstants.NoReg; l++)
                                {
                                    Add(new RegionRegionParameterKey(p.Name, i, l), pp[i - 1, l - 1]);
                                };
                            };
                        };
                    };
                };
            }
            finally
            {
                Console.WriteLine();
            }
        }


        public Random _rand;

        const string c_normalDistributionName = "NormalDistribution";
        const string c_triangularDistributionName = "TriangularDistribution";
        const string c_exponentialDistributionName = "ExponentialDistribution";
        const string c_gammaDistributionName = "GammaDistribution";

        public Parameters() : this(null)
        {
        }

        public Parameters(Random rand)
        {
            UseReflectionToCreateObjects();
            _rand = rand;
        }

        public Parameters(int NumberOfYears, int NumberOfRegions, Random rand) : this(rand)
        {
        }

        public void SetToBestGuessValue()
        {           
            foreach (var p in _parameters.Values)
            {
                p.SetToBestGuessValue();
            }

            UseReflectionToSetFieldsCurrentValues();
        }

        public void SetToRandomValue()
        {
            foreach (var p in _parameters.Values)
            {
                p.SetToRandomValue();
            }

            UseReflectionToSetFieldsCurrentValues();
        }

        public void ReadExcelFile(string FileName)
        {
            if (FileName.EndsWith(".xls"))
                ReadExcel2003File(FileName);
            else if (FileName.EndsWith(".xlsm"))
                ReadExcel2007File(FileName);
            else
                throw new Exception("Unknown file format");
        }

        public void WriteCurrentValuesToFile(string FileName)
        {

            throw new NotImplementedException();

            //using (var i_writer = new StreamWriter(FileName))
            //{
            //    foreach (var p in _globalParameters.Values)
            //    {
            //        i_writer.Write(p.Key.Name);
            //        i_writer.Write(" = ");
            //        i_writer.WriteLine(p.Value);
            //    }

            //    foreach (var p in _regionParameters.Values)
            //    {
            //        i_writer.Write(p.Key.Name);
            //        i_writer.Write(" (" + ((RegionParameterKey)p.Key).RegionName + ")");
            //        i_writer.Write(" = ");
            //        i_writer.WriteLine(p.Value);
            //    }

            //    foreach (var p in _regionRegionParameters.Values)
            //    {
            //        i_writer.Write(p.Key.Name);
            //        i_writer.Write(" (" + ((RegionRegionParameterKey)p.Key).Region1 + ";" + ((RegionRegionParameterKey)p.Key).Region2 + ")");
            //        i_writer.Write(" = ");
            //        i_writer.WriteLine(p.Value);
            //    }

            //    foreach (var p in _yearParameters.Values)
            //    {
            //        i_writer.Write(p.Key.Name);
            //        i_writer.Write(" (" + Convert.ToString(((YearParameterKey)p.Key).Year) + ")");
            //        i_writer.Write(" = ");
            //        i_writer.WriteLine(p.Value);
            //    }

            //    foreach (var p in _yearRegionParameters.Values)
            //    {
            //        i_writer.Write(p.Key.Name);
            //        i_writer.Write(" (");
            //        i_writer.Write(Convert.ToString(((YearRegionParameterKey)p.Key).Year));
            //        i_writer.Write(";");
            //        i_writer.Write(((YearRegionParameterKey)p.Key).Region + ")");
            //        i_writer.Write(" = ");
            //        i_writer.WriteLine(p.Value);
            //    }
            //}
        }

        public void WriteValuesToCsv(int ScenarioId, int RunId, TextWriter GlobalCsv, TextWriter RegionCsv, TextWriter YearCsv, TextWriter RegionYearCsv)
        {
            if (GlobalCsv != null)
            {
                var globalParams = from p in _parameters.Values where p.Key is GlobalParameterKey select p;
                foreach (var p in globalParams)
                {
                    GlobalCsv.Write(ScenarioId);
                    GlobalCsv.Write(';');
                    GlobalCsv.Write(RunId);
                    GlobalCsv.Write(';');
                    GlobalCsv.Write(m_allParameterIdsByName[p.Key.Name]);
                    GlobalCsv.Write(';');
                    GlobalCsv.Write(p.Value);
                    GlobalCsv.WriteLine();
                }
            }

            if (RegionCsv != null)
            {
                var regionalParameters = from p in _parameters.Values where p.Key is RegionParameterKey select p;
                foreach (var p in regionalParameters)
                {
                    RegionCsv.Write(ScenarioId);
                    RegionCsv.Write(';');
                    RegionCsv.Write(RunId);
                    RegionCsv.Write(';');
                    RegionCsv.Write(m_allParameterIdsByName[p.Key.Name]);
                    RegionCsv.Write(';');
                    RegionCsv.Write(((RegionParameterKey)p.Key).RegionId);
                    RegionCsv.Write(';');
                    RegionCsv.Write(p.Value);
                    RegionCsv.WriteLine();
                }
            }

            // TODO: RegionRegionCsv

            if (YearCsv != null)
            {
                var yearlyParameters = from p in _parameters.Values where p.Key is YearParameterKey select p;
                foreach (var p in yearlyParameters)
                {
                    YearCsv.Write(ScenarioId);
                    YearCsv.Write(';');
                    YearCsv.Write(RunId);
                    YearCsv.Write(';');
                    YearCsv.Write(m_allParameterIdsByName[p.Key.Name]);
                    YearCsv.Write(';');
                    YearCsv.Write(((YearParameterKey)p.Key).Year + 1950);
                    YearCsv.Write(';');
                    YearCsv.Write(p.Value);
                    YearCsv.WriteLine();
                }
            }

            if (RegionYearCsv != null)
            {
                var yearRegionParameters = from p in _parameters.Values where p.Key is YearRegionParameterKey select p;
                foreach (var p in yearRegionParameters)
                {
                    RegionYearCsv.Write(ScenarioId);
                    RegionYearCsv.Write(';');
                    RegionYearCsv.Write(RunId);
                    RegionYearCsv.Write(';');
                    RegionYearCsv.Write(m_allParameterIdsByName[p.Key.Name]);
                    RegionYearCsv.Write(';');
                    RegionYearCsv.Write(((YearRegionParameterKey)p.Key).RegionId);
                    RegionYearCsv.Write(';');
                    RegionYearCsv.Write(((YearRegionParameterKey)p.Key).Year + 1950);
                    RegionYearCsv.Write(';');
                    RegionYearCsv.Write(p.Value);
                    RegionYearCsv.WriteLine();
                }
            }
        }

        public double GetItemByKey(ParameterKey key)
        {
            return _parameters[key].Value;
        }

        public IEnumerable<Parameter> GetAllParameters()
        {
            foreach (var p in _parameters.Values)
            {
                yield return p;
            }
        }

    }
}