﻿// Climate Framework for Uncertainty, Negotiation and Distribution
// Version 3.5
// (c) David Anthoff and Richard S.J. Tol
// http://www.fund-model.org
using System.Collections;

namespace Fund
{

    // Class that holds all output from one model run
    public class ModelOutput
    {
        private Damages m_damages = new Fund.Damages();
        private Incomes m_incomes = new Fund.Incomes();
        private Populations m_populations = new Fund.Populations();


        public Damages Damages { get { return m_damages; } set { m_damages = value; } }
        public Incomes Incomes { get { return m_incomes; } set { m_incomes = value; } }
        public Populations Populations { get { return m_populations; } set { m_populations = value; } }
    }

    // The damage in one year, region and sector in US$
    public class Damage
    {
        private int m_year;
        private int m_region;
        private Sector m_sector;
        private double m_damage;

        public Damage(int YearData, int RegionData, Sector SectorData, double DamageData)
        {
            m_year = YearData;
            m_region = RegionData;
            m_sector = SectorData;
            m_damage = DamageData;
        }

        public Damage()
        {
        }

        public int Year { get { return m_year; } set { m_year = value; } }
        public int Region { get { return m_region; } set { m_region = value; } }
        public Sector Sector { get { return m_sector; } set { m_sector = value; } }
        public double DamageValue { get { return m_damage; } set { m_damage = value; } }
    }

    // Collection of all Damages for one model run
    public class Damages
    {
        private ArrayList m_list;
        private Damage[, ,] m_damages = new Damage[LegacyConstants.NYear, LegacyConstants.NoReg, LegacyConstants.NoSector];

        public int Count { get { return GetCount(); } }
        public Damage this[int Index] { get { return GetItem(Index); } }

        public void Add(Damage Damage)
        {
            m_list.Add(Damage);
            m_damages[Damage.Year - 1, Damage.Region - 1, (int)Damage.Sector] = Damage;
        }

        public static Damages CalculateMarginalDamage(Damages Damages1, Damages Damages2)
        {
            Damage i_marginalDamage;
            Damages i_marginalDamages;
            Damage i_damage1, i_damage2;

            i_marginalDamages = new Damages();

            for (int i = 0; i < Damages1.Count; i++)
            {
                i_damage1 = Damages1[i];
                i_damage2 = Damages2.GetByYearRegionSector(i_damage1.Year, i_damage1.Region, i_damage1.Sector);

                i_marginalDamage = new Damage();
                i_marginalDamage.Year = i_damage1.Year;
                i_marginalDamage.Region = i_damage1.Region;
                i_marginalDamage.Sector = i_damage1.Sector;
                i_marginalDamage.DamageValue = (i_damage2.DamageValue - i_damage1.DamageValue) / 10000000;
                i_marginalDamages.Add(i_marginalDamage);
            }
            return i_marginalDamages;
        }

        public Damages()
        {
            m_list = new ArrayList(LegacyConstants.NYear * LegacyConstants.NoReg * LegacyConstants.NoSector);
        }

        public Damage GetByYearRegionSector(int Year, int Region, Sector Sector)
        {
            return m_damages[Year - 1, Region - 1, (int)Sector];
        }

        public int GetCount()
        {
            return m_list.Count;
        }

        public Damage GetItem(int Index)
        {
            return (Damage)m_list[Index];
        }
    }

    // GDP for each year and region in US$
    public class Incomes
    {
        private double[,] m_incomes = new double[LegacyConstants.NYear, LegacyConstants.NoReg];
        public double this[int Year, int Region] { get { return GetItem(Year, Region); } }

        public void Add(int Year, int Region, double Income)
        {
            m_incomes[Year - 1, Region - 1] = Income;
        }

        public Incomes()
        {
        }

        public double GetItem(int Year, int Region)
        {
            return m_incomes[Year - 1, Region - 1];
        }

    }

    // Population for each year and region in absolute numbers
    public class Populations
    {
        private double[,] m_populations = new double[LegacyConstants.NYear, LegacyConstants.NoReg];
        public double this[int Year, int Region] { get { return GetItem(Year, Region); } }

        public void Add(int Year, int Region, double Population)
        {
            m_populations[Year - 1, Region - 1] = Population;
        }

        public Populations()
        {
        }

        public double GetItem(int Year, int Region)
        {
            return m_populations[Year - 1, Region - 1];
        }

    }

}