﻿// Climate Framework for Uncertainty, Negotiation and Distribution
// Version 3.5
// (c) David Anthoff and Richard S.J. Tol
// http://www.fund-model.org
using System;
namespace Fund
{

    public static class LegacyConstants
    {
        public const int NYear = 1049;      // number of years
        public const int NoReg = 16;       // number of regions
        public const int NoSector = 15;    // number of sectors
        public const double DBsT = 0.04;     // base case yearly warming
    }

    // Enum that can be used to categorise damage. If the number of options here
    // is changed, make sure to also adjust NoSector!
    public enum Sector
    {
        sWater,
        sForests,
        sHeating,
        sCooling,
        sAgriculture,
        sDryland,
        sSeaProtection,
        sImigration,
        sSpecies,
        sDeath,
        sMorbidity,
        sWetland,
        sEmigration,
        sHurrican,
        sExtratropicalStorms
    }

    public class FundParameter : Parameters
    {
        [Parameter]
        public double consleak;

        // n2o uncertainty
        [Parameter]
        public double n2oadd;

        // ch4 uncertainty
        [Parameter]
        public double ch4add;

        // life ch4
        [Parameter]
        public double lifech4;

        // life n20
        [Parameter]
        public double lifen2o;

        // life sea
        [Parameter]
        public double lifesea;

        // share of agriculture
        [Parameter]
        public RegionParameter agrish0;

        // base regional temperature parameters
        [Parameter]
        public RegionParameter bregtmp;

        [Parameter]
        public RegionParameter temp90;

        // Extra methane emissions for Megan's extreme scenarios
        [Parameter]
        public YearParameter RunAwayCH4;

        // income growth per year
        // economic growth
        [Parameter]
        public YearRegionParameter ypcgrowth;

        // population growth per year
        [Parameter]
        public YearRegionParameter pgrowth;

        [Parameter]
        public RegionParameter gdp90;

        [Parameter]
        public RegionParameter pop90;

        // emission from deforestation, MtCO2
        [Parameter(Name = "ForestEmm")]
        public YearRegionParameter forestemm;

        // autonomous energy efficiency improvement
        [Parameter(Name = "AEEI")]
        public YearRegionParameter aeei;

        // autonomous carbon efficiency improvements
        [Parameter(Name = "ACEI")]
        public YearRegionParameter acei;

        // exogenous methane emissions
        [Parameter]
        public YearRegionParameter ch4em;

        // benchmark water resources costs
        [Parameter]
        public RegionParameter wrbm;

        // benchmark heating costs
        [Parameter]
        public RegionParameter hebm;

        // benchmark cooling costs
        [Parameter]
        public RegionParameter cebm;

        // hurricane damage
        [Parameter]
        public RegionParameter hurrbasedam;

        // income elasticity
        [Parameter]
        public double hurrdamel;

        // hurricane casualties
        [Parameter]
        public RegionParameter hurrbasedead;

        // income elasticity
        [Parameter]
        public double hurrdeadel;

        // climate sensitivity
        [Parameter]
        public double hurrpar;

        // climate non-linearity
        [Parameter]
        public double hurrnl;

        [Parameter]
        public double extratropicalstormsdamel;

        [Parameter]
        public RegionParameter extratropicalstormspar;

        [Parameter]
        public double extratropicalstormsdeadel;

        [Parameter]
        public RegionParameter extratropicalstormsbasedam;

        [Parameter]
        public RegionParameter extratropicalstormsbasedead;

        // benchmark forestry costs
        [Parameter]
        public RegionParameter forbm;

        // benchmark dengue fever
        [Parameter]
        public RegionParameter dfbs;

        [Parameter]
        public RegionParameter dfch;

        // benchmark schistosomiasis
        [Parameter]
        public RegionParameter smbs;

        [Parameter]
        public RegionParameter smch;

        // benchmark malaria
        [Parameter]
        public RegionParameter malbs;

        [Parameter]
        public RegionParameter malch;

        // life-losses from cardheat.
        [Parameter]
        public RegionParameter chmlbm;

        [Parameter]
        public RegionParameter chmqbm;

        [Parameter]
        public RegionParameter chplbm;

        [Parameter]
        public RegionParameter chpqbm;

        //life-losses from cardcold
        [Parameter]
        public RegionParameter ccmlbm;

        [Parameter]
        public RegionParameter ccmqbm;

        [Parameter]
        public RegionParameter ccplbm;

        [Parameter]
        public RegionParameter ccpqbm;

        // life-losses from resp
        [Parameter]
        public RegionParameter rlbm;

        [Parameter]
        public RegionParameter rqbm;

        // wetland losses sea level rise
        [Parameter]
        public RegionParameter wlbm;

        [Parameter]
        public RegionParameter wmbm;

        [Parameter]
        public RegionParameter wetmax;

        // dryland losses sea level rise
        [Parameter]
        public RegionParameter dlbm;

        // protection costs sea level rise 
        [Parameter]
        public RegionParameter coast;

        [Parameter]
        public RegionParameter pc;

        // agricultural losses
        [Parameter]
        public RegionParameter agrbm;

        [Parameter]
        public RegionParameter aglbm;

        [Parameter]
        public RegionParameter agopt;

        [Parameter]
        public RegionParameter agtime;

        [Parameter(Name = "agcbm")]
        public RegionParameter agcbmInternal;

        // uncertainty about scenarios
        [Parameter]
        public RegionParameter pgadd;

        [Parameter]
        public RegionParameter ecgradd;

        [Parameter]
        public RegionParameter aeeiadd;

        [Parameter]
        public RegionParameter aceiadd;

        [Parameter]
        public RegionParameter foremadd;

        // benchmark dengue fever
        // Not uncertain, because they are data
        [Parameter]
        public RegionParameter diamort;

        [Parameter]
        public RegionParameter diayld;

        // reduction cost parameter for forestry
        // forestry cost constant
        [Parameter]
        public RegionParameter forconst;

        // forestry cost linear
        [Parameter]
        public RegionParameter forlinear;

        // forestry cost square
        [Parameter]
        public RegionParameter forsquare;

        // initial income
        [Parameter]
        public RegionParameter gdp0;

        // initial SO2
        [Parameter]
        public RegionParameter so20;

        // methane parameters
        [Parameter]
        public RegionParameter ch4par1;

        [Parameter]
        public RegionParameter ch4par2;

        // exogenous n2o emissions
        [Parameter]
        public YearRegionParameter n2oem;

        // nitrous oxide parameters
        [Parameter]
        public RegionParameter n2opar1;

        [Parameter]
        public RegionParameter n2opar2;

        // SF6 emissions, 1990
        [Parameter]
        public RegionParameter sf60;

        // SF6 parameters
        [Parameter]
        public double sf6ypc;

        [Parameter]
        public double sf6gdp;

        [Parameter]
        public double lifesf6;

        [Parameter]
        public double roebakercsparameter;

        [Parameter]
        public double sf6add;

        // exogenous temperatures 
        [Parameter]
        public YearRegionParameter scentemp;

        // dead-to-years-of-life-disabled transformation parameters
        [Parameter]
        public RegionParameter d2dm;

        [Parameter]
        public RegionParameter d2ds;

        [Parameter]
        public RegionParameter d2dd;

        [Parameter]
        public RegionParameter d2dc;

        [Parameter]
        public RegionParameter d2dr;

        // dead-to-years-of-life-lost transformation parameters
        [Parameter]
        public RegionParameter d2lm;

        [Parameter]
        public RegionParameter d2ls;

        [Parameter]
        public RegionParameter d2ld;

        [Parameter]
        public RegionParameter d2lc;

        [Parameter]
        public RegionParameter d2lr;

        [Parameter]
        public RegionParameter slrprtp;

        // initial population
        [Parameter]
        public RegionParameter pop0;

        [Parameter]
        public double bioloss;

        [Parameter]
        public double biosens;

        [Parameter]
        public double bioshare;

        [Parameter]
        public double ClimateSensitivity;

        [Parameter]
        public double seas;  // modal sea level rise T=1

        [Parameter]
        public double watechrate;

        [Parameter]
        public double forel;

        [Parameter]
        public double fornl;

        [Parameter]
        public double wrel;

        [Parameter]
        public double wrnl;

        [Parameter]
        public double wrpl;

        [Parameter]
        public double heel;

        [Parameter]
        public double henl;

        [Parameter]
        public double ceel;

        [Parameter]
        public double cenl;

        [Parameter]
        public double agnl;

        [Parameter]
        public double agel;

        [Parameter]
        public double malnl;

        [Parameter]
        public double dfnl;

        [Parameter]
        public double smnl;

        [Parameter]
        public double forco2;

        [Parameter]
        public double plusel;

        [Parameter]
        public double spbm;

        [Parameter]
        public double vslbm;

        [Parameter]
        public double vslel;

        [Parameter]
        public double vmorbbm;

        [Parameter]
        public double vmorbel;

        [Parameter]
        public double vmorbypc0;

        [Parameter]
        public double vslypc0;

        [Parameter]
        public double vyll;

        [Parameter]
        public double diamortel;

        [Parameter]
        public double diamortnl;

        [Parameter]
        public double diayldel;

        [Parameter]
        public double diayldnl;

        [Parameter]
        public double emcst;

        [Parameter]
        public double immcst;

        [Parameter]
        public double cvlin;

        [Parameter]
        public double rlin;

        [Parameter]
        public double maxcardvasc;

        [Parameter]
        public double vbel;

        [Parameter]
        public double lifeco1;

        [Parameter]
        public double lifeco2;

        [Parameter]
        public double lifeco3;

        [Parameter]
        public double lifeco4;

        [Parameter]
        public double lifeco5;

        [Parameter]
        public double valbase;

        [Parameter]
        public double slrwvypc0;

        [Parameter]
        public RegionParameter valinc;

        [Parameter]
        public double wvbm;

        [Parameter]
        public RegionParameter wetland90;

        [Parameter]
        public double dvbm;

        [Parameter]
        public double dvydl;

        [Parameter]
        public double wvel;

        [Parameter]
        public double slrwvpopdens0;

        [Parameter]
        public double wvpdl;

        [Parameter]
        public double wvsl;

        [Parameter]
        public RegionParameter taxmp;

        [Parameter]
        public double DBsTa;

        [Parameter]
        public double nospecbase;

        // Jiehan's prtp
        [Parameter]
        public RegionParameter prtp;

        // Which gas is used for the marginal
        // evaluation (0=CO2, 1=CH4, 2=N20)
        [Parameter(Name = "marginalgas")]
        public double marginalGas;

        // In which year is the emission
        // of the marginal gases done?
        [Parameter(Name = "emissionperiod")]
        public double emissionPeriod;

        [Parameter(Name = "Migrate")]
        public RegionRegionParameter migrate;

        [Parameter]
        public double co2pre;

        [Parameter]
        public RegionParameter emissint0;

        [Parameter]
        public RegionParameter urbcorr;

        // pop dens at coast, relative to total
        [Parameter]
        public RegionParameter coastpd;

        [Parameter]
        public RegionParameter plus90;

        [Parameter]
        public RegionParameter cardvasc90;

        [Parameter]
        public RegionParameter resp90;

        [Parameter]
        public double incdens;

        [Parameter]
        public double extratropicalstormsnl;

        [Parameter]
        public RegionParameter area0;

        [Parameter]
        public double so2pop;

        [Parameter]
        public double so2inc;

        [Parameter]
        public double so2carb;

        [Parameter]
        public double knowpar;

        [Parameter]
        public double knowgpar;

        [Parameter]
        public double gwpch4;

        [Parameter]
        public double gwpn2o;

        [Parameter]
        public double TaxConstant;

        [Parameter]
        public double TaxEmInt;

        [Parameter]
        public double TaxThreshold;

        [Parameter]
        public double TaxDepreciation;

        [Parameter]
        public double MaxCostFall;

        [Parameter]
        public double TerrCO2Sens;

        [Parameter]
        public double TerrCO2Stock0;

        [Parameter]
        public double LifeTempConst;

        [Parameter]
        public double LifeTempLin;

        [Parameter]
        public double timeofuncertaintystart;

        public FundParameter(int NumberOfYears, int NumberOfRegions, Random rand)
            : base(NumberOfYears, NumberOfRegions, rand)
        {
        }
    }
}